/*
 * renum.h
 *
 * Contains generic NFORenum stuff
 *
 * Copyright 2004-2006 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _RENUM_RENUM_H_INCLUDED_
#define _RENUM_RENUM_H_INCLUDED_

#include"getopt.h"

//*************************
//    OS SPECIFIC STUFF
//*************************

/* file handling defs:
* dirname contains the name of the sprites directory -- caps on Windows, lowercase on *nix
* foo_ext contains the extention to use for foo -- caps on Windows, lowercase on *nix
*/
#define dirname "sprites/"

#ifdef _WIN32
    #define nfo_ext ".NFO"
    #define new_ext ".NEW"
    #define bak_ext ".BAK"
#else
    #define nfo_ext ".nfo"
    #define new_ext ".new"
    #define bak_ext ".bak"
#endif

//*************************
//  END OS SPECIFIC STUFF
//*************************

typedef unsigned int uint;
typedef unsigned char uchar;

inline int _FORCE_INT_(int x){return x;}

enum{EOK,EWARN=3,EERROR,EPARSE,EFILE,EDATA,EFATAL};
void SetCode(int);

#define INTERNAL_ERROR(type,func,var,val)\
	if(true){\
        IssueMessage(0,INTERNAL_##type##_ERROR,__FILE__,__LINE__,#func,#var,_FORCE_INT_(val));\
		assert(false);\
        exit(EFATAL);\
	}else

#define DEFAULT(func,var)\
    default:\
        INTERNAL_ERROR(SWITCH,func,var,var);

#define VERIFY(cond,func,var)\
    if(!(cond))INTERNAL_ERROR(LOGIC,func,var,var);\
    else

#define EXPECTED_BYTES(off) (off>>24)
#define EXPECTED_LOC(off) (off&0xFFFFFF)

#define SINGLETON(class)\
public:\
    static class&Instance(){static class obj;return obj;}\
    static const class&CInstance(){return Instance();}\
private:\
    class();\
    class(const class&);\
    void operator=(const class&);

#define spritenum() (GetState(DIFF)?-1:(int)_spritenum)

#endif//_RENUM_RENUM_H_INCLUDED_
